/*
    liban - base function to do usefull things related to simulation and scientific work
    Copyright (C) 1999-2005 Stephane Magnenat <nct@ysagoon.com>
    Copyright (c) 2004-2005 Antoine Beyeler <antoine.beyeler@epfl.ch>
    Copyright (C) 2005 Laboratory of Intelligent Systems, EPFL, Lausanne
    See AUTHORS for details

    This program is free software. You can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __AN_FORMULA_H
#define __AN_FORMULA_H

#include <string>
#include <vector>
#include <stack>
#include <map>
#include <typeinfo>

/*!	\file Formula.h
	\brief Forumla, a formule parser, bytecompiler and interpreter
*/

namespace An
{
	//! A Formula
	class Formula
	{
	public:
		//! Result returned by Formula::parse, including error description
		struct ParseResult
		{
			//! Result
			enum Result
			{
				SUCCESS = 0, //!< parsing was a success
				UNDEFINED_VARIABLE, //!< some variable are used but not defined
				SYNTAX_ERROR //!< a syntax error has occured
			} result;
			//! error position in formula text
			size_t errorPos;
			
			//! Constructor, set to success
			ParseResult() { result = SUCCESS; errorPos = 0; }
			//! Cast to bool, return true on success
			operator bool () const { return result == SUCCESS; }
		};
		
		//! Result stack
		typedef std::stack<double> ResultStack;
		
		//! An token from the bytecode, can be a variable or an operation
		struct BytecodeToken
		{
			//! evaluate and return results on stack
			virtual bool eval(ResultStack &stack) = 0;
			//! return the name of the token
			virtual const char *getType(void) { return typeid(*this).name(); }
		};
		
		//! pointer to a unary function
		typedef double (*UnaryFunctionPtr)(double);
		
		//! The Bytecode is a vector of tokens
		typedef std::vector<BytecodeToken *> Bytecode;
		//! The VariableMap is a map of string to pointers to double
		typedef std::map<std::string, const double *> VariableMap;
		//! The FunctionMap is a map of string to pointers to function.
		typedef std::map<std::string, UnaryFunctionPtr> FunctionMap;
		
	protected:
		//! text of formula
		std::string formulaText;
		//! map of variable name to variable values
		VariableMap variableMap;
		//! map of function name to functions
		FunctionMap functionMap;
		//! bytecode of the formula
		Bytecode bytecode;
		//! has the formula been parsed?
		bool parsed;
		
	public:
		//! Constructor, take a text in input
		Formula(const std::string &formula, bool useStandardFunctions = true);
		//! Destructor, free tokens
		virtual ~Formula();
		//! Set a variable from the formula to an external value
		void setVariable(const std::string &name, const double &value);
		//! Add an unary function
		void addUnaryFunction(const std::string &name, UnaryFunctionPtr funcPtr);
		//! Parse, return a ParseResult struct indicating if the parsing was successfull, and if not, the error. All formula variables must be bound to C++ ones by calls to setVariable before this calling this method.
		ParseResult parse(void);
		//! Evaluate the formula
		double eval(void);
		
	protected:
		// standard functions
		//! f(x) = x*x
		static double sqr(double v) { return v*v; }
		//! f(x) = 1 if x > 0; -1 if x < 0; 0 otherwise
		static double sgn(double v) { return v > 0.0 ? 1.0 : (v < 0.0 ? -1.0 : 0.0); }
	};
}

#endif
